"""
The two-dimension detector generates images of the reflected intensity.
The purpose of the Image class stored in this module is the investigation
and manipulation of these images.
"""


import numpy as np


class Image:
    """
    This class stores information about the detector images.

    Attributes:
        file_path (:py:attr:`str`):
            File path for the image.
        array (:py:attr:`array_like`):
            The image described as an array.
        array_original (:py:attr:`array_like`):
            The original value of the image array when it was loaded from disk.
        array_e (:py:attr:`array_like`):
            The errors on each pixel of the array.
        bkg (:py:attr:`float`):
            The background that was subtracted from the image.
        bkg_e (:py:attr:`float`):
            The uncertainty on the background.

    Args:
        file_path (:py:attr:`str`): The file path for the image.
        data (:py:class:`pandas.DataFrame`, optional): Experimental data about the measurement. Defaults to :py:attr:`None`.
        transpose (:py:attr:`bool`, optional): Should the data be rotated by 90 degrees? Defaults to :py:attr:`False`.
    """

    def __init__(self, array: np.ndarray, transpose: bool = False):
        """
        Initialisation of the :py:class:`islatu.image.Image` class, includes
        assigning uncertainties.
        """
        if transpose:
            array = array.T
        self.array = array
        self.array_original = np.copy(array)
        self.array_e = self.initial_std_devs
        self.bkg = 0
        self.bkg_e = 0

    @property
    def nominal_values(self):
        """
        Get the nominal values of the image array.

        Returns:
            :py:attr:`array_like`: Nominal values of image.
        """
        return self.array

    @property
    def initial_std_devs(self):
        """
        Get the standard deviation values of the original raw image array.

        Returns:
            :py:attr:`array_like`: Standard deviation values of image.
        """
        array_error = np.sqrt(self.array_original)
        array_error[np.where(self.array_original == 0)] = 1
        return array_error

    @property
    def shape(self):
        """
        Array shape

        Returns:
            :py:attr:`tuple` of :py:attr:`int`: The shape of the image.
        """
        return self.array.shape

    def __repr__(self):
        """
        Custom representation.

        Returns:
            :py:attr:`array_like`: Image array.
        """
        return self.array

    def __str__(self):
        """
        Custom string.

        Returns:
            :py:attr:`array_like`: Image array.
        """
        return self.array

    def crop(self, crop_function, **kwargs):
        """
        Perform an image crop based on some function.

        Args:
            crop_function (:py:attr:`callable`): The function to crop the data.
            **kwargs (:py:attr:`dict`): The crop function keyword arguments.
        """
        self.array = crop_function(self.array, **kwargs)
        self.array_e = crop_function(self.array_e, **kwargs)

    def background_subtraction(self, background_subtraction_function,
                               **kwargs):
        """
        Perform a background subtraction based on some function.

        Args:
            background_subtraction_function (:py:attr:`callable`): The function to model the data and therefore remove the background.
            **kwargs (:py:attr:`dict`): The background substraction function keyword arguments.
        """

        bkg_sub_info = background_subtraction_function(
            self, **kwargs
        )
        # Store the calculated background, and its error.
        self.bkg, self.bkg_e = bkg_sub_info.bkg, bkg_sub_info.bkg_e

        # Do the subtraction.
        self.array = self.array - self.bkg
        self.array_e = np.sqrt(self.bkg_e**2 + self.array_e**2)

        # Expose information relating to the background subtraction for
        # meta-analyses.
        return bkg_sub_info

    def sum(self):
        """
        Perform a summation on the image's array.

        Returns:
            A tuple taking the form (summed_intensity, summed_intensity_e).
        """
        intensity = np.sum(self.array)
        intensity_e = np.sqrt(np.sum(self.array_e**2))

        return intensity, intensity_e



"""
This module contains:

Parsing functions used to extract information from experimental files.

Classes used to help make parsing more modular. These include the NexusBase
class and its children.
"""

# We've gotta access the _value attribute on some NXobjects.
# pylint: disable=protected-access


import json
import os
from typing import List
from abc import abstractmethod
from json import JSONDecodeError

import nexusformat.nexus.tree as nx
from nexusformat.nexus import nxload
import pandas as pd
import numpy as np
import h5py


from islatu.scan import Scan2D
from islatu.image import Image
from islatu.data import Data
from islatu.region import Region
from islatu.debug import debug
from islatu.metadata import Metadata


class NexusBase(Metadata):
    """
    This class contains *mostly* beamline agnostic nexus parsing convenience
    stuff. It's worth noting that this class still makes a series of assumptions
    about how data is laid out in a nexus file that can be broken. Instead of
    striving for some impossible perfection, this class is practical in its
    assumptions of how data is laid out in a .nxs file, and will raise if an
    assumption is violated. All instrument-specific assumptions that one must
    inevitably make to extract truly meaningful information from a nexus file
    are made in children of this class.

    Attrs:
        file_path:
            The local path to the file on the local filesystem.
        nxfile:
            The object produced by loading the file at file_path with nxload.
    """

    def __init__(self, local_path: str):
        super().__init__(local_path)
        self.nxfile = nxload(local_path)

    @property
    def src_path(self):
        """
        The name of this nexus file, as it was recorded when the nexus file was
        written.
        """
        return self.nxfile.file_name

    @property
    def detector(self):
        """
        Returns the NXdetector instance stored in this NexusFile.

        Raises:
            ValueError if more than one NXdetector is found.
        """
        det, = self.instrument.NXdetector
        return det

    @property
    def instrument(self):
        """
        Returns the NXinstrument instanced stored in this NexusFile.

        Raises:
            ValueError if more than one NXinstrument is found.
        """
        instrument, = self.entry.NXinstrument
        return instrument

    @property
    def entry(self) -> nx.NXentry:
        """
        Returns this nexusfile's entry.

        Raises:
            ValueError if more than one entry is found.
        """
        entry, = self.nxfile.NXentry
        return entry

    @property
    def default_signal(self) -> np.ndarray:
        """
        The numpy array of intensities pointed to by the signal attribute in the
        nexus file.
        """
        return self.default_nxdata[self.default_signal_name].nxdata

    @property
    def default_axis(self) -> np.ndarray:
        """
        Returns the nxdata associated with the default axis.
        """
        return self.default_nxdata[self.default_axis_name].nxdata

    @property
    def default_signal_name(self):
        """
        Returns the name of the default signal.
        """
        return self.default_nxdata.signal

    @property
    def default_axis_name(self) -> str:
        """
        Returns the name of the default axis.
        """
        return self.entry[self.entry.default].axes

    @property
    def default_nxdata_name(self):
        """
        Returns the name of the default nxdata.
        """
        return self.entry.default

    @property
    def default_nxdata(self) -> np.ndarray:
        """
        Returns the default NXdata.
        """
        return self.entry[self.default_nxdata_name]

    # A hack to tell pylint that this class is still meant to be abstract.
    @property
    @abstractmethod
    def default_axis_type(self) -> str:
        return super().default_axis_type()


class I07Nexus(NexusBase):
    """
    This class extends NexusBase with methods useful for scraping information
    from nexus files produced at the I07 beamline at Diamond.
    """
    excalibur_detector_2021 = "excroi"
    excalibur_04_2022 = "exr"
    pilatus_02_2024 = "PILATUS"
    excalibur_08_2024= "EXCALIBUR"

    @property
    def local_data_path(self) -> str:
        """
        The local path to the data (.h5) file. Note that this isn't in the
        NexusBase class because it need not be reasonably expected to point at a
        .h5 file.

        Raises:
            FileNotFoundError if the data file cant be found.
        """
        file = _try_to_find_files(
            [self._src_data_path[0]], [self.local_path])[0]
        return file

    @property
    def detector_name(self) -> str:
        """
        Returns the name of the detector that we're using. Because life sucks,
        this is a function of time.
        """
        if "excroi" in self.entry:
            return I07Nexus.excalibur_detector_2021
        if "exr" in self.entry:
            return I07Nexus.excalibur_04_2022
        if "PILATUS" in self.entry:
            return I07Nexus.pilatus_02_2024
        if "EXCALIBUR" in self.entry:
            return I07Nexus.excalibur_08_2024
        # Couldn't recognise the detector.
        raise NotImplementedError()

    @property
    def default_axis_name(self) -> str:
        """
        Returns the name of the default axis.
        """
        return self.entry[self.entry.default].axes
    
    @property
    def default_axis_type(self) -> str:
        """
        Returns the type of our default axis, either being 'q', 'th' or 'tth'.
        """
        if self.default_axis_name == 'qdcd':
            return 'q'
        if self.default_axis_name == 'diff1chi':
            return 'th'  
        if self.default_axis_name == 'diff1delta':
            return 'tth'

        
        
        #add in option for EH2 scanning alpha
        if self.default_axis_name == 'diff2alpha_value_set':
            return 'th'
        
        #add in option for EH1 scanning chi
        if self.default_axis_name == 'diff1chi_value_set':
            return 'th'
    
        # It's also possible that self.default_axis_name isn't recorded in some
        # nexus files. Just in case, let's check the length of diff1delta.
        if isinstance(self.instrument["diff1delta"].value.nxdata, np.ndarray):
            return 'tth'

    def _get_ith_region(self, i: int):
        """
        Returns the ith region of interest found in the .nxs file.

        Args:
            i:
                The region of interest number to return. This number should
                match the ROI name as found in the .nxs file (generally not 0
                indexed).

        Returns:
            The ith region of interest found in the .nxs file.
        """
        x_1 = self.detector[self._get_region_bounds_key(i, 'x_1')][0]
        x_2 = self.detector[self._get_region_bounds_key(i, 'Width')][0] + x_1
        y_1 = self.detector[self._get_region_bounds_key(i, 'y_1')][0]
        y_2 = self.detector[self._get_region_bounds_key(i, 'Height')][0] + y_1
        return Region(x_1, x_2, y_1, y_2)

    def get_string_for_json(self,instrument_string):
        try:
            json_str = self.instrument[
                instrument_string]._value.decode("utf-8")
        except AttributeError:
            json_str = self.instrument[
                instrument_string]._value
        return json_str
    
    def region_from_json(self,json_str):
        json_str = json_str.replace('u', '')
        json_str = json_str.replace("'", '"')
        try:
            roi_dict = json.loads(json_str)
            return [Region.from_dict(roi_dict['Region_1'])]
        except JSONDecodeError:
            debug.log("region list from nexus file not in valid JSON format")
            return []
    @property
    def signal_regions(self) -> List[Region]:
        """
        Returns a list of region objects that define the location of the signal.
        Currently there is nothing better to do than assume that this is a list
        of length 1.
        """
        if self.detector_name == I07Nexus.excalibur_detector_2021:
            return [self._get_ith_region(i=1)]
        excaliburlist=[I07Nexus.excalibur_04_2022,I07Nexus.excalibur_08_2024]

        if self.detector_name in excaliburlist:
            json_str=self.get_string_for_json("ex_rois/excalibur_ROIs")
            return self.region_from_json(json_str)

        if self.detector_name== I07Nexus.pilatus_02_2024:
            json_str=self.get_string_for_json("p3_rois/pilatus3_ROIs")
            return self.region_from_json(json_str)

        raise NotImplementedError()

    @property
    def background_regions(self) -> List[Region]:
        """
        Returns a list of region objects that define the location of background.
        Currently we just ignore the zeroth region and call the rest of them
        background regions.
        """
        if self.detector_name == I07Nexus.excalibur_detector_2021:
            return [self._get_ith_region(i)
                    for i in range(2, self._number_of_regions+1)]
        excaliburlist=[I07Nexus.excalibur_04_2022,I07Nexus.excalibur_08_2024]
        if self.detector_name in excaliburlist:
            # Make sure our code executes for bytes and strings.
            try:
                json_str = self.instrument[
                    "ex_rois/excalibur_ROIs"]._value.decode("utf-8")
            except AttributeError:
                json_str = self.instrument[
                    "ex_rois/excalibur_ROIs"]._value
            # This is badly formatted and cant be loaded by the json lib. We
            # need to make a series of modifications.
            json_str = json_str.replace('u', '')
            json_str = json_str.replace("'", '"')

            roi_dict = json.loads(json_str)
            bkg_roi_list = list(roi_dict.values())[1:2]
            return [Region.from_dict(x) for x in bkg_roi_list]
        #add in similar option for pilatus option
        if self.detector_name== I07Nexus.pilatus_02_2024:
            try: 
                json_str=self.instrument["p3_rois/pilatus3_ROIs"]._value.decode("utf-8")
            except AttributeError:
                json_str=self.instrument["p3_rois/pilatus3_ROIs"]._value
            json_str = json_str.replace('u', '')
            json_str = json_str.replace("'", '"')
            
            roi_dict = json.loads(json_str)
            bkg_roi_list = list(roi_dict.values())[1:2]
            return [Region.from_dict(x) for x in bkg_roi_list]
        raise NotImplementedError()

    @property
    def probe_energy(self):
        """
        Returns the energy of the probe particle parsed from this NexusFile.
        """
        return float(self.instrument.dcm1energy.value)

    @property
    def transmission(self):
        """
        Proportional to the fraction of probe particles allowed by an attenuator
        to strike the sample.
        """
        if 'filterset' in self.instrument:
            valsarr=np.array(self.instrument.filterset.transmission)
            if len(np.shape(valsarr))==0:
                outarr=np.array([self.instrument.filterset.transmission])
            else:
                outarr=valsarr
            return outarr
        elif 'fatt' in self.instrument:
            valsarr=np.array(self.instrument.fatt.transmission)
            if len(np.shape(valsarr))==0:
                outarr=np.array([self.instrument.fatt.transmission])
            else:
                outarr=valsarr
            return outarr
        else:
            debug.log(f"\n No transmission value found in expected location, set transmission to 1 \n")
            return np.array([float(1)])

    @property
    def detector_distance(self):
        """
        Returns the distance between sample and detector.
        """
        return float(self.instrument.diff1detdist.value)

    @property
    def _src_data_path(self):
        """
        Returns the raw path to the data file. This is useless if you aren't on
        site, but used by islatu to guess where you've stored the data file
        locally.
        """
        # This is far from ideal; there currently seems to be no standard way
        # to refer to point at information stored outside of the nexus file.
        # If you're a human, it's easy enough to find, but with code this is
        # a pretty rubbish task. Here I just grab the first .h5 file I find
        # and run with it.
        found_h5_files = []
        datanxfilepaths=[]

        def recurse_over_nxgroups(nx_object, found_h5_files,datanxfilepaths):
            """
            Recursively looks for nxgroups in nx_object that, when cast to a
            string, end in .h5.
            """
            for key in nx_object:
                new_obj = nx_object[key]
                if key == "data":
                    if isinstance(new_obj,nx.NXlink):
                        found_h5_files.append(new_obj.nxfilename)
                        datanxfilepaths.append(new_obj.nxfilepath)
                    # if new_obj.tree[8:-9].endswith(".h5"):
                    #     found_h5_files.append(new_obj.tree[8:-9])
                if str(new_obj).endswith(".h5"):
                    found_h5_files.append(str(new_obj))
                if str(new_obj).endswith(".h5['/data']"):
                    found_h5_files.append(str(new_obj)[:-9])
                if isinstance(new_obj, nx.NXgroup):
                    recurse_over_nxgroups(new_obj, found_h5_files,datanxfilepaths)

        recurse_over_nxgroups(self.nxfile, found_h5_files,datanxfilepaths)

        return np.array([found_h5_files[0],datanxfilepaths[0]])

    @property
    def _region_keys(self) -> List[str]:
        """
        Parses all of the detector's dictionary keys and returns all keys
        relating to regions of interest.
        """
        return [key for key in self.detector.keys() if key.startswith("Region")]

    @property
    def _number_of_regions(self) -> int:
        """
        Returns the number of regions of interest described by this nexus file.
        This *assumes* that the region keys take the form f'region_{an_int}'.
        """
        split_keys = [key.split('_') for key in self._region_keys]

        return max([int(split_key[1]) for split_key in split_keys])

    def _get_region_bounds_key(self, region_no: int, kind: str) -> List[str]:
        """
        Returns the detector key relating to the bounds of the region of
        interest corresponding to region_no.

        Args:
            region_no:
                An integer corresponding the the particular region of interest we're interested in generating a key for.
            kind:
                The kind of region bounds keys we're interested in. This can take the values:
                    'x_1', 'width', 'y_1', 'height'
                where '1' can be replaced with 'start' and with/without caps on first letter of width/height.

        Raises:
            ValueError if 'kind' argument is not one of the above.

        Returns:
            A list of region bounds keys that is ordered by region number.
        """
        # Note that the x, y swapping is a quirk of the nexus standard, and is
        # related to which axis on the detector varies most rapidly in memory.
        if kind in ('x_1', 'x_start'):
            insert = 'X'
        elif kind in ('width', 'Width'):
            insert = 'Width'
        elif kind in ('y_1', 'y_start'):
            insert = 'Y'
        elif kind in ('height', 'Height'):
            insert = 'Height'
        else:
            raise ValueError(
                "Didn't recognise 'kind' argument.")

        return f"Region_{region_no}_{insert}"


def i07_dat_to_dict_dataframe(file_path):
    """
    Parses a .dat file recorded by I07, returning a [now mostly obsolete] tuple
    containing a metadata dictionary and a pandas dataframe of the data.

    Though outdated, this is still a handy way to parse the DCD normalization
    .dat file.

    Args:
        (:py:attr:`str`): The ``.dat`` file to be read.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`dict`: The metadata from the ``.dat`` file.
            - :py:class:`pandas.DataFrame`: The data from the ``.dat`` file.
    """
    f_open = open(file_path, "r", encoding='utf-8')
    # Neither the data nor the metadata are being read yet.
    data_reading = False
    metadata_reading = False

    # Create the dictionaries to be populated.
    data_dict = {}
    metadata_dict = {}
    # Create the list to be filled with lists for each line
    data_lines = []

    for line in f_open:
        # This string incidates the start of the metadata.
        if "<MetaDataAtStart>" in line:
            metadata_reading = True
        # This string indicates the end of the metadata.
        if "</MetaDataAtStart>" in line:
            metadata_reading = False
        # This string indicates the start of the data.
        if " &END" in line:
            data_reading = True
            # Set counter to minus two, such that when is
            # reaches the data it is 0.
            count = -2
        # When the metadata section is being read populate the metadata_dict
        if metadata_reading:
            if "=" in line:
                metadata_in_line = []
                for i in line.split("=")[1:]:
                    try:
                        j = float(i)
                    except ValueError:
                        j = i
                    metadata_in_line.append(j)
                metadata_dict[line.split("=")[0]] = metadata_in_line
        # When the data section is being read, make the list of the zeroth line
        # the titles and everything after is the data_lines list of lists.
        if data_reading:
            count += 1
            if count == 0:
                titles = line.split()
            if count > 0:
                data_lines.append(line.split())
    f_open.close()
    # Sort the data_lines list of lists to transpore and make into a dict where
    # the keys are the titles.
    for j, _ in enumerate(data_lines[0]):
        list_to_add = []
        for i, _ in enumerate(data_lines):
            try:
                list_to_add.append(float(data_lines[i][j]))
            except ValueError:
                list_to_add.append(data_lines[i][j])
        count = 0
        if j >= len(titles):
            data_dict[str(count)] = list_to_add
            count += 1
        else:
            data_dict[titles[j]] = list_to_add
    return metadata_dict, pd.DataFrame(data_dict)


def load_images_from_h5(h5_file_path,datanxfilepath,transpose=False):
    """
    Loads images from a .h5 file.

    Args:
        h5_file_path:
            Path to the h5 file from which we're loading images.
        transpose:
            Should we take the transpose of these images? Defaults to True.
    """
    internal_data_path = datanxfilepath#'data'
    images = []
    debug.log("Loading images from file " + h5_file_path, unimportance=0)
    with h5py.File(h5_file_path, "r") as file_handle:
        dataset = file_handle[internal_data_path]#[()]

        num_images = dataset.shape[0]
        # Prepare to show a progress bar for image loading.
        debug.log(f"Loading {num_images} images.", unimportance=2)
        for i in range(num_images):
            debug.log("Currently loaded " + str(i+1) + " images.",  end="\r")
            images.append(Image(dataset[i], transpose=transpose))
        # This line is necessary to prevent overwriting due to end="\r".
        debug.log("")
        debug.log(f"Loaded all {num_images} images.", unimportance=2)

    return images


def i07_nxs_parser(file_path: str,remove_indices=None,adjustments=None):
    """
    Parses a .nxs file acquired from the I07 beamline at diamond, returning an
    instance of Scan2D. This process involves loading the images contained in
    the .h5 file pointed at by the .nxs file, as well as retrieving the metadata
    from the .nxs file that is relevant for XRR reduction.

    Args:
        file_path:
            Path to the .nxs file.

    Returns:
        An initialized Scan2D object containing all loaded detector frames, as
        well as the relevant metadata from the .nxs file.
    """
    # Use the magical parser class that does everything for us.
    i07_nxs = I07Nexus(file_path)
    detname=i07_nxs.detector_name
    if 'attenuation_filters_moving' in i07_nxs.entry[f'{detname}'].keys():
        try:
            attenuationvalues=i07_nxs.entry[f'{detname}/attenuation_value'].nxdata
            movingfilters=[0 if attenuationvalues[i]==attenuationvalues[i-1] else 1 for i in np.arange(1,len(attenuationvalues[0:5]))]
        except (AttributeError,TypeError):
            debug.log("unable to read in attenuation information, possible missing attenuation h5 file. Will assume no moving attenuation.", unimportance=2)
            attenuationvalues=i07_nxs.entry[f'{detname}/attenuation_value']
            movingfilters=[]
        remove_indices=np.where(movingfilters)[0]
        #remove_indices=np.where(np.array(i07_nxs.entry[f'{detname}/attenuation_filters_moving']))[0]
        remove_indices+=1
        
    # Load the images, taking a transpose if necessary (because which axis is
    # x and which is why is determined by fast vs slow detector axes in memory).
    if i07_nxs.detector_name in [
            I07Nexus.excalibur_detector_2021,
            I07Nexus.excalibur_04_2022,
            I07Nexus.pilatus_02_2024,
            I07Nexus.excalibur_08_2024]:
        images = load_images_from_h5(i07_nxs.local_data_path,i07_nxs._src_data_path[1], transpose=False)

    # The dependent variable.
    rough_intensity = i07_nxs.default_signal
    rough_intensity_e = np.sqrt(rough_intensity)



    # The independent variable.
    axis = i07_nxs.default_axis
    axis_type = i07_nxs.default_axis_type
    theta_offset=0
    q_offset=0

    if hasattr(adjustments,'new_axis_name'):
        axis_name=adjustments.new_axis_name
        if hasattr(i07_nxs.instrument[f'{axis_name}'],"value_set"):
            axis=i07_nxs.instrument[f'{axis_name}'].value_set.nxdata
        else:
            axis=i07_nxs.instrument[f'{axis_name}'].value.nxdata
    if hasattr(adjustments,'new_axis_type'):
        axis_type=adjustments.new_axis_type
    if hasattr(adjustments,'theta_offset'):
        theta_offset=adjustments.theta_offset
    if hasattr(adjustments,'q_offset'):
        q_offset=adjustments.q_offset

    if axis_type== 'q':
        data = Data(rough_intensity, rough_intensity_e, i07_nxs.probe_energy,
                    q_vectors=axis-q_offset)
    elif axis_type  == 'th':
        data = Data(rough_intensity, rough_intensity_e, i07_nxs.probe_energy,
                    theta=axis-theta_offset)
    elif axis_type  == 'tth':
        data = Data(rough_intensity, rough_intensity_e, i07_nxs.probe_energy,                
                    theta=(axis/2-theta_offset))
    else:
        raise NotImplementedError(
            f"{axis_type} is not a supported axis type. Axis name={axis_name}")


    # Returns the Scan2D object
    return Scan2D(data, i07_nxs, images, remove_indices)


def _try_to_find_files(filenames: List[str],
                       additional_search_paths: List[str]):
    """
    Check that data files exist if the file parsed by parser pointed to a
    separate file containing intensity information. If the intensity data
    file could not be found in its original location, check a series of
    probable locations for the data file. If the data file is found in one
    of these locations, update file's entry in self.data.

    Returns:
        :py:attr:`list` of :py:attr:`str`:
            List of the corrected, actual paths to the files.
    """
    found_files = []

    # If we had only one file, make a list out of it.
    if not hasattr(filenames, "__iter__"):
        filenames = [filenames]

    cwd = os.getcwd()
    start_dirs = [
        cwd,  # maybe file is stored near the current working dir
        # To search additional directories, add them in here manually.
    ]
    start_dirs.extend(additional_search_paths)

    local_start_directories = [x.replace('\\', '/') for x in start_dirs]
    num_start_directories = len(local_start_directories)

    # Now extend the additional search paths.
    for i in range(num_start_directories):
        search_path = local_start_directories[i]
        split_srch_path = search_path.split('/')
        for j in range(len(split_srch_path)):
            extra_path_list = split_srch_path[:-(j+1)]
            extra_path = '/'.join(extra_path_list)
            local_start_directories.append(extra_path)

    # This line allows for a loading bar to show as we check the file.
    for i, _ in enumerate(filenames):
        # Better to be safe... Note: windows is happy with / even though it
        # defaults to \
        filenames[i] = str(filenames[i]).replace('\\', '/')

        # Maybe we can see the file in its original storage location?
        if os.path.isfile(filenames[i]):
            found_files.append(filenames[i])
            continue

        # If not, maybe it's stored locally? If the file was stored at
        # location /a1/a2/.../aN/file originally, for a local directory LD,
        # check locations LD/aj/aj+1/.../aN for all j<N and all LD's of
        # interest. This algorithm is a generalization of Andrew McCluskey's
        # original approach.

        # now generate a list of all directories that we'd like to check
        candidate_paths = []
        split_file_path = str(filenames[i]).split('/')
        for j in range(len(split_file_path)):
            local_guess = '/'.join(split_file_path[j:])
            for start_dir in local_start_directories:
                candidate_paths.append(
                    os.path.join(start_dir, local_guess))

        # Iterate over each of the candidate paths to see if any of them contain
        # the data file we're looking for.
        found_file = False
        for candidate_path in candidate_paths:
            if os.path.isfile(candidate_path):
                # File found - add the correct file location to found_files
                found_files.append(candidate_path)
                found_file = not found_file
                debug.log("Data file found at " + candidate_path + ".")
                break

        # If we didn't find the file, tell the user.
        if not found_file:
            raise FileNotFoundError(
                "The data file with the name " + filenames[i] + " could "
                "not be found. The following paths were searched:\n" +
                "\n".join(candidate_paths)
            )
    return found_files



"""
This module contains both the Data class and the MeasurementBase class.
In a reflectometry measurement, the experimental data corresponds to the
reflected intensity as a function of scattering vector Q. In a typical
diffractometer, Q is a virtual axis, calculated geometrically from various motor
positions. The Data class takes care of these conversions, exposing q, theta,
intensity, reflectivity, and energy.

The MeasurementBase class defines a simple class that is Data, but that also has
metadata.
"""

import numpy as np
from scipy.constants import physical_constants


class Data:
    """
        The base class of all Islatu objects that contain data.

        Attributes:
            intensity:
                A numpy array containing intensities in this dataset.
            intensity_e:
                A numpy array containing the corresponding errors in intensity.
            theta:
                A numpy array containing the probe particle's angle of
                incidence at each intensity.
            q_vectors:
                A numpy array containing the magnitude of the probe particle's
                scattering vector for each intensity value.
            energy:
                The energy of the probe particle used to acquire this data. This
                is necessary to swap between theta and q.

        Args:
            intensity:
                A numpy array of the intensities in this dataset.
            intensity_e:
                The errors on the intensities.
            energy:
                The energy of the probe particle used to acquire this data.
            theta:
                A numpy array containing the probe particle's angle of
                incidence at each intensity. NOTE: only one of theta/q needs to
                be provided.
            q_vectors:
                A numpy array containing the magnitude of the probe particle's
                scattering vector for each intensity value. NOTE: only one of
                theta/q needs to be provided.
        """

    def __init__(self, intensity, intensity_e, energy, theta=None,
                 q_vectors=None):

        self.intensity = intensity
        self.intensity_e = intensity_e
        self.energy = energy

        if (theta is None) and (q_vectors is None):
            raise ValueError(
                "Either theta or q must be provided to create a Data instance"
            )

        # When using properties, it wont matter which of these ends up as None.
        self._theta = theta
        self._q = q_vectors

    @property
    def reflectivity(self) -> np.array:
        """
        Returns the intensity, normalized such that the maximum value of the
        intensity is equal to 1. To acquire
        """
        return self.intensity/np.amax(self.intensity)

    @property
    def reflectivity_e(self) -> np.array:
        """
        Returns the errors on the intensity, divided by the maximum value of the
        intensity array.
        """
        return self.intensity_e/np.amax(self.intensity)
    @property
    def reflectivity_nonorm(self) -> np.array:
        """
        Returns the intensity, unnormalised
        """
        return self.intensity

    @property
    def reflectivity_e_nonorm(self) -> np.array:
        """
        Returns the errors on the intensity, unnormalised
        """
        return self.intensity_e

    @property
    def q_vectors(self) -> np.array:
        """
        Returns self._q if this instance of Data was generated from q-data.
        Otherwise, converts from self._theta to q.
        """
        if (self._q is None) and (self._theta is not None):
            return self._theta_to_q(self._theta, self.energy)
        else:
            return self._q

    @q_vectors.setter
    def q_vectors(self, value) -> None:
        """
        Sets self._q.
        """
        self._q = value

    @property
    def theta(self) -> np.array:
        """
        Returns self._theta if this instance of Data was generate from th-data.
        Otherwise, converts from scattered q to theta.
        """
        if (self._theta is None) and (self._q is not None):
            return self._q_to_theta(self._q, self.energy)
        else:
            return self._theta

    @theta.setter
    def theta(self, value) -> None:
        self._theta = value

    def _theta_to_q(self, theta, energy) -> np.array:
        """
        Calculates the scattering vector Q from diffractometer theta.

        Args:
            theta (:py:attr:`str`):
                Array of theta values to be converted.
            energy (:py:attr:`float`):
                Energy of the incident probe particle.
        """
        planck = physical_constants["Planck constant in eV s"][0] * 1e-3
        speed_of_light = physical_constants[
            "speed of light in vacuum"][0] * 1e10
        q_values = np.sin(np.radians(theta)) / (planck * speed_of_light)

        q_values *= energy * 4.0 * np.pi
        return q_values

    def _q_to_theta(self, q_values, energy) -> np.array:
        """
        Calculates the diffractometer theta from scattering vector Q.

        Args:
            theta (:py:attr:`str`):
                Array of theta values to be converted.
            energy (:py:attr:`float`):
                Energy of the incident probe particle.
        """
        planck = physical_constants["Planck constant in eV s"][0] * 1e-3
        speed_of_light = physical_constants[
            "speed of light in vacuum"][0] * 1e10
        theta_values = planck * speed_of_light * \
            np.arcsin(q_values / (energy * 4 * np.pi))

        theta_values = theta_values*180/np.pi

        return theta_values

    def remove_data_points(self, indices):
        """
        Convenience method for the removal of a specific data point by its
        index.

        Args:
            indices:
                The indices to be removed.
        """
        if self._q is not None:
            self._q = np.delete(self._q, indices)
        if self._theta is not None:
            self._theta = np.delete(self._theta, indices)

        self.intensity = np.delete(self.intensity, indices)
        self.intensity_e = np.delete(self.intensity_e, indices)


class MeasurementBase(Data):
    """
    All measurements derive from this class.

    Attrs:
        metadata:
            The metadata relevant to this measurement.
    """

    def __init__(self, intensity, intensity_e, energy, metadata, theta=None,
                 q=None) -> None:
        # Initialize the Data.
        super().__init__(intensity, intensity_e, energy, theta, q)
        # Store the metadata.
        self.metadata = metadata



"""
As reflectometry measurements typically consist of multiple scans at different
attenutation, we must stitch these together.
"""

from typing import List

import numpy as np

from islatu.scan import Scan


def concatenate(scan_list: List[Scan]):
    """
    Concatenate each of the datasets together.

    Args:
        scans:
            List of reflectometry scans.

    Returns:
        :py:attr:`tuple`: Containing:
            - q-values.
            - Reflected intensities.
            – Errors on reflected intensities.
    """

    q_vectors = np.array([])
    intensity = np.array([])
    intensity_e = np.array([])
    for scan in scan_list:
        q_vectors = np.append(q_vectors, scan.q_vectors)
        intensity = np.append(intensity, scan.intensity)
        intensity_e = np.append(intensity_e, scan.intensity_e)
    return q_vectors, intensity, intensity_e


def rebin(q_vectors, reflected_intensity, new_q=None, rebin_as="linear",
          number_of_q_vectors=5000):
    """
    Rebin the data on a linear or logarithmic q-scale.

    Args:
        q_vectors:
            q - the current q vectors.
        reflected_intensity (:py:attr:`tuple`):
            (I, I_e) - The current reflected intensities, and their errors.
        new_q (:py:attr:`array_like`):
            Array of potential q-values. Defaults to :py:attr:`None`. If this
            argument is not specified, then the new q, R values are binned
            according to rebin_as and number_of_q_vectors.
        rebin_as (py:attr:`str`):
            String specifying how the data should be rebinned. Options are
            "linear" and "log". This is only used if the new_q are unspecified.
        number_of_q_vectors (:py:attr:`int`, optional):
            The max number of q-vectors to be using initially in the rebinning
            of the data. Defaults to :py:attr:`400`.

    Returns:
        :py:attr:`tuple`: Containing:
            - q: rebinned q-values.
            - intensity: rebinned intensities.
            - intensity_e: rebinned intensity errors.
    """

    # Unpack the arguments.
    q = q_vectors
    R, R_e = reflected_intensity

    # Required so that logspace/linspace encapsulates the whole data.
    epsilon = 0.001

    if new_q is None:
        # Our new q vectors have not been specified, so we should generate some.
        if rebin_as == "log":
            new_q = np.logspace(
                np.log10(q[0]),
                np.log10(q[-1] + epsilon), number_of_q_vectors)
        elif rebin_as == "linear":
            new_q = np.linspace(q.min(), q.max() + epsilon,
                                number_of_q_vectors)

    binned_q = np.zeros_like(new_q)
    binned_R = np.zeros_like(new_q)
    binned_R_e = np.zeros_like(new_q)

    for i in range(len(new_q)-1):
        indices = []
        inverse_var = []
        for j in range(len(q)):
            if new_q[i] <= q[j] < new_q[i + 1]:
                indices.append(j)
                inverse_var.append(1/float(R_e[j]**2))

        # Don't bother doing maths if there were no recorded q-values between
        # the two bin points we were looking at.
        if len(indices) == 0:
            continue

        # We will be using inverse-variance weighting to minimize the variance
        # of the weighted mean.
        sum_of_inverse_var = np.sum(inverse_var)

        # If we measured multiple qs between these bin locations, then average
        # the data, weighting by inverse variance.
        for j in indices:
            binned_R[i] += R[j]/(R_e[j]**2)
            binned_q[i] += q[j]/(R_e[j]**2)

        # Divide by the sum of the weights.
        binned_R[i] /= sum_of_inverse_var
        binned_q[i] /= sum_of_inverse_var

        # The stddev of an inverse variance weighted mean is always:
        binned_R_e[i] = np.sqrt(1/sum_of_inverse_var)

    # Get rid of any empty, unused elements of the array.
    cleaned_q = np.delete(binned_q, np.argwhere(binned_R == 0))
    cleaned_R = np.delete(binned_R, np.argwhere(binned_R == 0))
    cleaned_R_e = np.delete(binned_R_e, np.argwhere(binned_R == 0))

    return cleaned_q, cleaned_R, cleaned_R_e



"""
Islatu's simple Debug class.
"""

DEFAULT_LOG_LEVEL = 1


class Debug:
    """
    A simple logger.

    Attrs:
        logging_level:
            Current logging level. Higher means more unimportant messages will
            be shown.
    """

    def __init__(self, logging_level):
        self.logging_level = logging_level

    def log(self, log_string, unimportance: int = 1, **kwargs):
        """
        Prints to stdout if self.logging_level >= unimportance.

        Args:
            log_string:
                The string to be printed.
            unimportance:
                A measure of unimportance assigned to the printing of this
                string. Very unimportant messages require a larger logging
                level to be printed. Defaults to 1.
        """
        if self.logging_level >= unimportance:
            print(log_string, **kwargs)


debug = Debug(DEFAULT_LOG_LEVEL)



"""
This module defines the Region object, whose instances define regions of
interest in images.
"""


class Region:
    """
    Instances of this class define regions of interest.
    """

    def __init__(self, x_start, x_end, y_start, y_end):

        # Make sure that x_end > x_start, etc.
        if x_end < x_start:
            x_start, x_end = x_end, x_start
        if y_end < y_start:
            y_start, y_end = y_end, y_start

        # These may be recorded as types other than int, but we really want
        # these to be integers so they can be used to index objects.
        self.x_start = int(x_start)
        self.x_end = int(x_end)
        self.y_start = int(y_start)
        self.y_end = int(y_end)

    @property
    def x_length(self):
        """
        Returns the length of the region in the x-direction.
        """
        return self.x_end - self.x_start

    @property
    def y_length(self):
        """
        Returns the length of the region in the y-direction.
        """
        return self.y_end - self.y_start

    @property
    def num_pixels(self):
        """
        returns the number of pixels in the region.
        """
        return self.x_length * self.y_length

    @classmethod
    def from_dict(cls, region_dict: dict):
        """
        Instantiates a Region from a dictionary with keys in:
            ['x', 'y', 'width', 'height'].

        This is to help loading dictionarys that are generated by calling
        json.loads on the NXcollections found in I07 nexus files as of
        27/04/2022.
        """
        x_start = int(region_dict['x'])
        y_start = int(region_dict['y'])
        x_end = x_start + int(region_dict['width'])
        y_end = y_start + int(region_dict['height'])
        return cls(x_start, x_end, y_start, y_end)

    def __eq__(self, other):
        """
        Allows for equality checks to be made between instances of Region.
        """
        if not isinstance(other, Region):
            return False

        return self.x_start == other.x_start and self.x_end == other.x_end \
            and self.y_start == other.y_start and self.y_end == other.y_end

    def __str__(self):
        return f"x_start: {self.x_start}, x_end: {self.x_end}, " + \
            f"y_start: {self.y_start}, y_end: {self.y_end}."



"""
This module contains functions whose purpose is simply to use the islatu
library to process data acquired from a specific instrument.
"""

from dataclasses import dataclass
from typing import List
from os import path
import os
from datetime import datetime
from ast import literal_eval as make_tuple
from types import SimpleNamespace
import os
from pathlib import Path
import subprocess
import time
import re
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
from yaml import load, dump
import numpy as np

from islatu.debug import debug
import islatu
import islatu.background as background
import islatu.corrections as corrections
import islatu.cropping as cropping
import islatu.io as io
from islatu.region import Region
from islatu.io import i07_dat_to_dict_dataframe
from islatu.refl_profile import Profile
from islatu.debug import debug
from islatu.config_loader import check_config_schema


# This could be done by reflection, but it feels slightly less arcane to use
# this kind of function map. It also gives these scripts a little more
# flexibility.
function_map = {
    'roi_subtraction': background.roi_subtraction,
    'None': None,
    'i07': io.i07_nxs_parser,
    'crop': cropping.crop_to_region
}


@dataclass
class Creator:
    """
    Simple dataclass to store information relating to the person that created
    this dataset.
    """
    name: str = 'Unknown'
    affiliation: str = 'Unknown'
    time: datetime = datetime.now()


@dataclass
class Origin:
    """
    Simple dataclass to store information relating to the experiment.
    """
    contact: str = 'My local contact'
    facility: str = 'Diamond Light Source'
    id: str = None
    title: str = None
    directory_path: str = None
    date: str = str(datetime.now())
    year: str = None


@dataclass
class Measurement:
    """
    This dataclass stores measurement-specific metadata.
    """
    scheme: str = 'q-dispersive'
    q_range: List[str] = (str(-np.inf), str(np.inf))
    theta_axis_name: str = 'dcdtheta'
    q_axis_name: str = 'qdcd'
    transpose: bool = False
    qz_dimension: int = 1
    qxy_dimension: int = 0


@dataclass
class Experiment:
    """
    This dataclass stores more instrument-specific metadata.
    """
    instrument: str = 'i07'
    probe: str = 'x-ray'
    energy: float = 12.5
    measurement: Measurement = Measurement()
    sample: str = None


class DataSource:
    """
    This class stores information relating both to the experiment, and to the
    data processor.
    """

    def __init__(self, title, origin=Origin(), experiment=Experiment(),
                 links=None):
        self.origin = origin
        self.origin.title = title
        self.experiment = experiment
        self.links = links


@dataclass
class Software:
    """
    This dataclass stores information relating to the software used to carry
    out the any reduction/processing steps (in this case, islatu of course).
    """
    name: str = 'islatu'
    link: str = 'https://islatu.readthedocs.io'
    version: str = islatu.__version__


@dataclass
class DataState:
    """
    This class stores more reduction specific parameters.
    """

    background = None
    resolution = None
    dcd = None
    transmission = None
    intensity = None
    rebinned = None

@dataclass
class ProcessArgs:
    """
    a class to hold all information about the processing settings requested, and provide parsing checks, then job submission    
    """

    data_path: str
    yaml_path: str
    version_path: str
    python_version: str
    cluster: str | None = None
    lower_bound: int | None = None
    upper_bound: int | None = None
    scan_numbers: list[int] | None = None
    output: str | None = None
    limit_q: list[str] = None
    verbose: int = 0
    islatufolder= str = None
    jobfile_template: str = None
    jobfile_name: str = None



    def resolve_defaults(self):
        """
        cast all string paths into proper path objects
        """
        self.islatufolder = Path(self.islatufolder or Path.home() / "islatu")
        self.version_path = Path(self.version_path)
        self.data_path = Path(self.data_path)
        self.yaml_path = Path(self.yaml_path)
        self.output = Path(self.output or self.data_path / "processed")
        self.jobfile_template = Path(self.jobfile_template or self.version_path / "template.sh")
        self.jobfile_name = self.jobfile_name or "jobscript_local.sh"
        self.processing_path = self.data_path / "processing"

    def set_logging(self):

        if self.verbose is None:
            self.verbose = 0

        # Set islatu's logger to requested verbosity.
        debug.logging_lvl = self.verbose

    def parse_yaml(self):
        # If execution reaches here, we're processing the scan locally. First look
        # for the .yaml file if we weren't explicitly told where it is.
        if self.yaml_path is not None:
            return self
        debug.log("Searching for .yaml files in '" + self.data_path +
                    "' and '" + self.processing_path + "'.")

        # Search in both the processing directory and the data directory.
        files = []

        # Only check in the processing directory if it actually exists.
        if os.path.exists(self.processing_path):
            files.extend([self.processing_path + x
                            for x in os.listdir(self.processing_path)])

        # The data_path should definitely exist. If it doesn't, we shouldn't be
        # unhappy about an error being raised at this point.
        files.extend(os.listdir(self.data_path))

        # Work out which of these files are .yaml files.
        yaml_files = [x for x in files if x.endswith(".yaml")]
        debug.log(".yaml files found: " + str(yaml_files))

        # If we didn't find exactly one .yaml file, complain.
        if len(yaml_files) != 1:
            generic_err_str = (
                "Could not uniquely determine location of .yaml file.\n" +
                "Searched directories " + self.processing_path + " and " +
                self.data_path + ".\n" + "Hoped to find exactly one file, " +
                "but found " + str(len(yaml_files)) + ". "
            )
            if len(yaml_files) > 1:
                generic_err_str += "Names of found files are: " + \
                    str(yaml_files) + "."
            raise FileNotFoundError(generic_err_str)
        else:
            # We only found one .yaml, so that's our guy.
            self.yaml_path = yaml_files[0]

    def parse_scan_numbers(self):

        # Default to smallest possible scan number (0).
        if self.lower_bound is None:
            self.lower_bound = 0

        # Make a number that will always be bigger than all other numbers.
        if self.upper_bound is None:
            self.upper_bound = float('inf')



        # If execution reaches here, we've successfully found the .yaml file.
        # Next lets try to work out what scan numbers are in the data directory if
        # we weren't told explicitly which scan numbers we should be looking at.
        if self.scan_numbers is None:
            debug.log(
                "Scan numbers not explicitly given. Searching for scans " +
                "in directory " + str(self.data_path) + "."
            )
            # Grab every valid looking nexus file in the directory.
            nexus_files = [x for x in os.listdir(
                self.data_path) if x.endswith(".nxs")]

            # Make noise if we didn't find any .nxs files.
            generic_cant_find_nxs = (
                "Couldn't find any nexus (.nxs) files in the data directory '" +
                str(self.data_path)
            )
            if len(nexus_files) == 0:
                raise FileNotFoundError(
                    generic_cant_find_nxs + "'."
                )

            # So, we found some .nxs files. Now lets grab the scan numbers from
            # these files.
            debug.log("Scans located: " + str(nexus_files))
            nexus_files = [int(x.replace(".nxs", '').replace("i07-", ''))
                        for x in nexus_files]

            # Now select the subset of these scan numbers that lies within the
            # closed interval [self.lower_bound, self.upper_bound].
            self.scan_numbers = [x for x in nexus_files if
                                x >= self.lower_bound and x <= self.upper_bound]
            self.scan_numbers.sort()
            debug.log("Scan numbers found: " + str(self.scan_numbers) + ".")

            # Make sure we found some scans.
            if len(self.scan_numbers) == 0:
                raise FileNotFoundError(
                    generic_cant_find_nxs +
                    " whose scan numbers were greater than or equal to " +
                    str(self.lower_bound) +
                    " and less than or equal to " + str(self.upper_bound) + "."
                )

    def parse_q_limits(self):
        if self.limit_q is not None:
            if len(self.limit_q) % 3 != 0:
                raise ValueError(
                    f"""
                    --limit_q must have a number of arguments passed to it that is
                    a multiple of three. Instead, {len(self.limit_q)} arguments were
                    found. Please use the pattern:
                        -L N1 qmin1 qmax1 N2 qmin2 qmax2 ...
                    where N1 is a scan number, qmin1 is the minimum q for the
                    scan with scan number N1, and qmax1 is the maximum acceptable q
                    for the scan with scan number N1, etc.. Please refer to the
                    --help for more information.
                    """
                )
            # Okay, this is presumably properly formatted. Lets turn this into a
            # list of dictionaries that we can pass directly to the
            # profile.subsample_q method.
            q_subsample_dicts = []
            for i, _ in enumerate(self.limit_q):
                if i % 3 == 0:
                    # We're on a new scan, so we'll need a new subsample dict.
                    q_subsample_dicts.append({})

                    # Now grab that dict we just created and give it our new scan
                    # index. Note that if i%3 != 0, then we can skip the creation
                    # of a new dictionary.
                    q_subsample_dicts[-1]['scan_ID'] = self.limit_q[i]
                elif i % 3 == 1:
                    # Convert every 2nd and 3rd value to a float – these will be
                    # our q limits.
                    self.limit_q[i] = float(self.limit_q[i])
                    q_subsample_dicts[-1]['q_min'] = self.limit_q[i]
                elif i % 3 == 2:
                    # Convert every 2nd and 3rd value to a float – these will be
                    # our q limits.
                    self.limit_q[i] = float(self.limit_q[i])
                    q_subsample_dicts[-1]['q_max'] = self.limit_q[i]
            self.limit_q = q_subsample_dicts

    def create_jobscript(self):
        self.islatufolder=f'{Path.home()}/islatu'
        if not os.path.exists(self.islatufolder):
            os.makedirs(self.islatufolder)
        i=1
        self.save_path=f'{self.islatufolder}/jobscript_{i}.py'
        while (os.path.exists(str(self.save_path))):
            i += 1
            save_file_name = f'{self.islatufolder}/jobscript_{i}.py'
            self.save_path = Path(save_file_name)
            if i > 1e7:
                raise ValueError(
                    "naming counter hit limit therefore exiting ")   
        with open(self.save_path,'x') as f:
            f.write("from islatu.runner import i07reduce\n")
            f.write(f"scans = {self.scan_numbers}\nyamlpath='{self.yaml_path}'\ndatapath='{self.data_path}'\noutfile='{self.output}'\nqsubdict={self.limit_q}\n")
            f.write("i07reduce(scans, yamlpath, datapath,filename=outfile, q_subsample_dicts=qsubdict)")
        #f.write(f"i07reduce({self.scan_numbers}, {self.yaml_path}, {self.data_path},\
        #      filename={self.output}, q_subsample_dicts={self.limit_q})")

    def create_jobfile(self):
        if self.jobfile_template is None:
            self.jobfile_template: str = Path(f'{self.version_path}/islatu/CLI/islatuscript_template.sh')
        if self.jobfile_name is None:
            self.jobfile_name: str = 'jobscript_local.sh'
        #load in template mapscript, new paths
        with open(self.jobfile_template) as f:
            lines=f.readlines()
        
        self.jobfile=f'{self.islatufolder}//{self.jobfile_name}'
        if os.path.exists(self.jobfile):
            f=open(self.jobfile,'w')
        else:
            f=open(self.jobfile,'x')
        save_path=self.save_path
        python_version=self.python_version
        for line in lines:
            phrase_matches=list(re.finditer(r'\${[^}]+\}',line))
            phrase_positions=[(match.start(),match.end()) for match in phrase_matches]
            outline=line
            
            for pos in phrase_positions:
                phrase=line[pos[0]:pos[1]]
                outphrase=phrase.strip('$').strip('{').strip('}')
                outline=outline.replace(phrase,str(locals()[f'{outphrase}']))
            f.write(outline)
        f.close()

    def check_slurmfiles(self):
        files=os.listdir(f'{Path.home()}/islatu')
        slurms=[x for x in files if '.out' in x]
        slurms.append(files[0])
        slurms.sort(key=lambda x: os.path.getmtime(f'{Path.home()}/islatu/{x}'))
        return slurms

    def run_cluster_job(self):

        self.create_jobscript()
        self.create_jobfile()

        startslurms=self.check_slurmfiles()
        endslurms=self.check_slurmfiles()
        count=0
        limit=0

        #call subprocess to submit job using wilson
        subprocess.run(["ssh","wilson",f"cd islatu \nsbatch {self.jobfile_name}"])
        while endslurms[-1]==startslurms[-1]:
            endslurms=self.check_slurmfiles()
            if count >50:
                limit=1
                break
            print(f'Job submitted, waiting for SLURM output.  Timer={5*count}',end="\r")
            time.sleep(5)
            count+=1
        if limit==1:
            print('Timer limit reached before new slurm ouput file found')
        else:
            print(f'Slurm output file: {Path.home()}/islatu//{endslurms[-1]}\n')
            breakerline='*'*35
            monitoring_line=f"\n{breakerline}\n ***STARTING TO MONITOR TAIL END OF FILE, TO EXIT THIS VIEW PRESS ANY LETTER FOLLOWED BY ENTER**** \n{breakerline} \n"
            print(monitoring_line)
            process = subprocess.Popen(["tail","-f",f"{Path.home()}/islatu//{endslurms[-1]}"], stdout=subprocess.PIPE, text=True)
            target_phrase="Reduced data stored"
            try:
                for line in process.stdout:
                    if "Loading images" in line:
                        print(line.strip(),'\n')
                    elif"Currently loaded" in line:
                        print(f"\r{line.strip()}", end='')
                    else:
                        print(line.strip())  # Print each line of output
                    if re.search(target_phrase, line):
                        print(f"Target phrase '{target_phrase}' found. Closing tail.")
                        break
                    if( "Errno" in line) or ("error" in line) or ("Error" in line):
                        print("error found. closing tail")
                        break
            finally:
                process.terminate()
                process.wait()
        print("Python script on cluster completed successfully")

    def parse_and_reduce(self):

        self.resolve_defaults()
        self.parse_yaml()
        self.parse_scan_numbers()
        self.parse_q_limits()
        if not self.cluster:
            i07reduce(self.scan_numbers, self.yaml_path, directory=self.data_path,
                filename=self.output, q_subsample_dicts=self.limit_q)
        else:
            self.run_cluster_job()


class Reduction:
    """
    This class contains all of the information pertaining to data reduction
    carried out on this reflectometry data.
    """

    def __init__(self, software=Software(), input_files=None,
                 data_state=DataState(), parser=io.i07_nxs_parser,
                 crop_function=cropping.crop_to_region, crop_kwargs=None,
                 bkg_function=background.fit_gaussian_1d, bkg_kwargs=None,
                 dcd_normalisation=None, sample_size=None, beam_width=None,
                 overwrite_transmission=None,normalisation=True,new_axis_name=None,new_axis_type=None):
        if input_files is None:
            input_files = []
        self.software = software
        self.input_files = input_files
        self.data_state = data_state
        self.parser = parser
        self.crop_function = crop_function
        self.crop_kwargs = crop_kwargs
        self.bkg_function = bkg_function
        self.bkg_kwargs = bkg_kwargs
        self.dcd_normalisation = dcd_normalisation
        self.sample_size = sample_size
        self.beam_width = beam_width
        self.overwrite_transmission=overwrite_transmission
        self.normalisation=normalisation
        self.new_axis_name=new_axis_name
        self.new_axis_type=new_axis_type


class Data:
    """
    This class stores information pertaining to the data collected in the
    experiment.
    """

    def __init__(self, columns=None, n_qvectors=50, q_min=None, q_max=None,
                 q_step=None, q_shape='linear'):
        if columns is None:
            columns = ['Qz / Aa^-1', 'RQz', 'sigma RQz, standard deviation',
                       'sigma Qz / Aa^-1, standard deviation']
        self.column_1 = columns[0]
        self.column_2 = columns[1]
        self.column_3 = columns[2]
        if len(columns) == 4:
            self.column_4 = columns[3]
        if columns == 'both':
            self.both = True
            self.column_4 = columns[3]
        self.rebin = True
        self.n_qvectors = n_qvectors
        self.q_min = q_min
        self.q_max = q_max
        self.q_step = q_step
        self.q_shape = q_shape


class Foreperson:
    """
    This class brings together all of the above classes and dataclasses into
    one big ball of yaml-able information.
    """

    def __init__(self, run_numbers, yaml_file, directory, title):
        self.creator = Creator()
        self.data_source = DataSource(title)
        self.reduction = Reduction()
        self.data = Data()
        self.yaml_file = yaml_file
        y_file = open(yaml_file, 'r', encoding='utf-8')
        recipe = load(y_file, Loader=Loader)
        y_file.close()

        self.setup(recipe)

        directory_path = directory.format(
            self.data_source.experiment.instrument,
            self.data_source.origin.year,
            self.data_source.origin.id)
        if path.isdir(directory_path):
            self.directory_path = directory_path
        else:
            raise FileNotFoundError(
                "The experiment directory <" + directory_path +
                "> cannot be found.")

        self.reduction.input_files = [
            self.directory_path + 'i07-' + str(r) + '.nxs' for r in run_numbers]

    def setup(self, recipe):
        """
        This is a McClusky special. I inherited it, and it works.
        Don't ask questions.
        """
        check_config_schema(recipe)

        keys = recipe.keys()
        # Populate information from the visit section
        if 'visit' in keys:
            self.data_source.origin.id = recipe['visit']['visit id']
            if 'date' in recipe['visit'].keys():
                self.data_source.origin.date = datetime.strptime(
                    str(recipe['visit']['date']), '%Y-%m-%d')
                self.data_source.origin.year = self.data_source.origin.date.year
            if 'local contact' in recipe['visit'].keys():
                self.data_source.origin.contact = recipe[
                    'visit']['local contact']
            if 'user' in recipe['visit'].keys():
                self.creator.name = recipe['visit']['user']
            if 'affiliation' in recipe['visit'].keys():
                self.creator.affiliation = recipe['visit']['user affiliation']
        else:
            raise ValueError(
                f"No visit given in {self.yaml_file}. " +
                "You must at least give a visit id")
        # Populate informatio from the information section
        if 'instrument' in keys:
            self.data_source.experiment.instrument = recipe['instrument']
            self.reduction.parser = function_map[recipe['instrument']]
        # Populate cropping information
        if 'crop' in keys:
            self.reduction.crop_function = function_map[
                recipe['crop']['method']]
            if 'kwargs' in recipe['crop']:
                self.reduction.crop_kwargs = recipe['crop']['kwargs']
        # Populate background subtraction method
        if 'background' in keys:
            self.reduction.bkg_function = function_map[
                recipe['background']['method']]
            if 'kwargs' in recipe['background']:
                self.reduction.bkg_kwargs = recipe['background']['kwargs']
        if 'transmission' in keys:
            self.reduction.overwrite_transmission=recipe['transmission']['values']
            
        if 'normalisation' in keys:
            self.reduction.normalisation=recipe['normalisation']['maxnorm']
            
        if 'adjustments'  in keys:
            if recipe['adjustments'] is not None:
                log_processing_stage("Setting adjustments")
                self.reduction.adjustments=SimpleNamespace(**recipe['adjustments'])
                for key,val in vars(self.reduction.adjustments).items():
                    debug.log(f'            {key} = {val}')
            # if 'new_axis_name' in recipe['adjustments'].keys():
            #     self.reduction.new_axis_name=recipe['adjustments']['new_axis_name']
            # if 'new_axis_type' in recipe['adjustments'].keys():
            #     self.reduction.new_axis_type=recipe['adjustments']['new_axis_type']
        # Populate the setup information
        if 'setup' in keys:
            if 'dcd normalisation' in recipe['setup'].keys():
                self.reduction.dcd_normalisation = recipe[
                    'setup']['dcd normalisation']
                self.data_source.links = {
                    'instrument reference': 'doi:10.1107/S0909049512009272'}
            if 'sample size' in recipe['setup'].keys():
                self.reduction.sample_size = make_tuple(recipe[
                    'setup']['sample size'])
                try:
                    _ = len(self.reduction.sample_size)
                    self.reduction.sample_size = self.reduction.sample_size[0]
                except TypeError:
                    pass
            else:
                raise ValueError("No sample size given in setup of {}.".format(
                    self.yaml_file))
            if 'beam width' in recipe['setup'].keys():
                self.reduction.beam_width = make_tuple(recipe[
                    'setup']['beam width'])
                try:
                    _ = len(self.reduction.beam_width)
                    self.reduction.beam_width = self.reduction.beam_width[0]
                except TypeError:
                    pass
            else:
                raise ValueError(
                    f"No beam width given in setup of {self.yaml_file}"
                )
            if 'theta axis' in recipe['setup'].keys():
                self.data_source.experiment.measurement.theta_axis_name = (
                    recipe['setup']['theta axis'])
            if 'q axis' in recipe['setup'].keys():
                self.data_source.experiment.measurement.q_axis_name = (
                    recipe['setup']['q axis'])
            if 'transpose' in recipe['setup'].keys():
                self.data_source.experiment.measurement.transpose = (
                    recipe['setup']['transpose'])
                if self.data_source.experiment.measurement.transpose:
                    self.data_source.experiment.measurement.qz_dimension = 0
                    self.data_source.experiment.measurement.qxy_dimension = 1
            if 'pixel max' in recipe['setup'].keys():
                self.data_source.experiment.measurement.pixel_max = recipe[
                    'setup']['pixel max']
            if 'hot pixel max' in recipe['setup'].keys():
                self.data_source.experiment.measurement.hot_pixel_max = recipe[
                    'setup']['hot pixel max']
        else:
            raise ValueError(f"No setup given in {self.yaml_file}.")
        if 'output_columns' in keys:
            if recipe['output_columns'] == 3:
                self.data = Data(
                    columns=[
                        'Qz / Aa^-1', 'RQz', 'sigma RQz, standard deviation'])
            if recipe['output_columns'] == 4:
                self.data = Data(columns='both')
        if 'rebin' in keys:
            if 'n qvectors' in recipe['rebin'].keys():
                self.data.n_qvectors = recipe['rebin']['n qvectors']
            elif 'min' in recipe['rebin'].keys() and 'max' in recipe[
                    'rebin'].keys() and 'step' in recipe['rebin'].keys():
                self.data.q_step = recipe['rebin']['step']
                if 'shape' in recipe['rebin'].keys():
                    self.data.q_shape = recipe['rebin']['shape']
            else:
                raise ValueError("Please define parameters of " +
                                 f"rebin in {self.yaml_file}.")
        else:
            self.data.rebin = False


def log_processing_stage(processing_stage):
    """
        Simple function to make logging slightly neater.
        """
    debug.log("-" * 10)
    debug.log(processing_stage, unimportance=0)
    debug.log("-" * 10)


def i07reduce(run_numbers, yaml_file, directory='/dls/{}/data/{}/{}/',
              title='Unknown', filename=None, q_subsample_dicts=None):
    """
    The runner that parses the yaml file and performs the data reduction.

    run_numbers (:py:attr:`list` of :py:attr:`int`):
        Reflectometry scans that make up the profile.
    yaml_file (:py:attr:`str`):
        File path to yaml config file
    directory (:py:attr:`str`):
        Outline for directory path.
    title (:py:attr:`str`):
        A title for the experiment.
    filename:
        Either a full path to the .dat file that will be produced by this
        function, or a directory. If a directory is given, then the filename
        will be automatically generated and the file will be placed in the
        specified directory.
    q_subsample_dicts:
        A list of dictionaries, which takes the form:
            [{'scan_ID': ID, 'q_min': q_min, 'q_max': q_max},...]
        where type(ID) = str, type(q_min)=float, type(q_max)=float.
    """

    # Make sure the directory is properly formatted.
    if not str(directory).endswith(os.sep):
        directory = str(directory) + os.sep
        the_boss = Foreperson(run_numbers, yaml_file, directory, title)

        # Necessary to distnguish the same data processed by different pipelines.
        yaml_pipeline_name = str(yaml_file).split(os.sep)[-1][:-5]

    files_to_reduce = the_boss.reduction.input_files

    log_processing_stage("File parsing")
    #return the_boss,files_to_reduce
    if hasattr(the_boss.reduction,"adjustments"):
        refl = Profile.fromfilenames(files_to_reduce, the_boss.reduction.parser,adjustments=the_boss.reduction.adjustments)
    else:
        refl = Profile.fromfilenames(files_to_reduce, the_boss.reduction.parser)

    # Set the energy correctly.
    the_boss.data_source.experiment.energy = refl.energy

    log_processing_stage("Cropping")
    # Currently, only crop_to_region is implemented.
    if the_boss.reduction.crop_function is not cropping.crop_to_region and \
            the_boss.reduction.crop_function is not None:
        raise NotImplementedError(
            "The only implemented cropping function is crop_to_region.")

    # Check to see if we were given an explicit cropping region. If not, use
    # the first (and likely only) signal region.
    if (the_boss.reduction.crop_function is cropping.crop_to_region and
            the_boss.reduction.crop_kwargs is None):
        try:
            roi = refl.scans[0].metadata.signal_regions[0]
        except IndexError:
            debug.log("No valid signal region found, regions will need to be defined in yaml settings file")
            return
        the_boss.reduction.crop_kwargs = {'region': roi}
        debug.log(f"Crop ROI '{str(roi)}' generated from the .nxs file.")
    elif 'x_end' in the_boss.reduction.crop_kwargs:
        the_boss.reduction.crop_kwargs = {
            'region': Region(**the_boss.reduction.crop_kwargs)
        }
    elif 'width' in the_boss.reduction.crop_kwargs:
        the_boss.reduction.crop_kwargs = {
            'region': Region.from_dict(region_dict=the_boss.reduction.crop_kwargs)
        }
    refl.crop(the_boss.reduction.crop_function,
              **the_boss.reduction.crop_kwargs)



    log_processing_stage("Subtracting background")
    # Before subtracting background, make sure that, by default, we're at least
    # trying to subtract background from roi_2.
    if the_boss.reduction.bkg_function is background.roi_subtraction:
        # Make sure we have the desired background regions.
        if the_boss.reduction.bkg_kwargs is None:
            the_boss.reduction.bkg_kwargs = {
                'list_of_regions': refl.scans[0].metadata.background_regions}
        elif 'x_end' in the_boss.reduction.bkg_kwargs:
            the_boss.reduction.bkg_kwargs = {
                'list_of_regions': Region(**the_boss.reduction.bkg_kwargs)
            }
        elif 'width' in the_boss.reduction.bkg_kwargs:
            the_boss.reduction.bkg_kwargs = {
                'list_of_regions': Region.from_dict(region_dict=the_boss.reduction.bkg_kwargs)
            }
        
        refl.bkg_sub(the_boss.reduction.bkg_function,
                      **the_boss.reduction.bkg_kwargs)
        the_boss.reduction.data_state.background = 'corrected'
    else:
        print("===== could not find suitable background settings so skipping step")
    



    log_processing_stage("Performing data corrections...")
    if the_boss.reduction.dcd_normalisation is not None:
        log_processing_stage("DCD normalisation")
        itp = corrections.get_interpolator(
            the_boss.reduction.dcd_normalisation, i07_dat_to_dict_dataframe)
        refl.qdcd_normalisation(itp)
        the_boss.reduction.data_state.dcd = 'normalised'
    


    log_processing_stage("Footprint correction.")
    refl.footprint_correction(
        the_boss.reduction.beam_width, the_boss.reduction.sample_size)
    



    log_processing_stage("Transmission normalisation.")
    if the_boss.reduction.overwrite_transmission is not None:
        overwrite_transmissions=the_boss.reduction.overwrite_transmission
    else:
        overwrite_transmissions=None

    refl.transmission_normalisation(overwrite_transmissions)
    the_boss.reduction.data_state.transmission = 'normalised'
    refl.concatenate()


    if q_subsample_dicts is not None:
        log_processing_stage(
            "Bounding q-vectors.")
        # We'll need to subsample a subset of our scans.
        for q_subsample_dict in q_subsample_dicts:
            refl.subsample_q(**q_subsample_dict)
        debug.log("Limited q-range on specified scans.")
    


    # Rebin the data, if the user requested this.
    if the_boss.data.rebin:
        log_processing_stage("Rebinning the data.")
        if the_boss.data.q_min is None:
            debug.log("Linearly rebinning data into " +
                      str(the_boss.data.n_qvectors) + " uniformly spaced " +
                      "points in q-space.", unimportance=2)
            refl.rebin(number_of_q_vectors=the_boss.data.n_qvectors)
        else:
            if the_boss.data.q_shape == 'linear':
                debug.log("Rebinning data linearly.", unimportance=2)
                spacing = np.linspace
            elif the_boss.data.q_shape == 'log':
                debug.log("Rebinning data logarithmically", unimportance=2)
                spacing = np.logspace
            debug.log(
                f"Spacing generated from {refl.q_vectors.min()}Å to " +
                f"{refl.q_vectors.max()}Å.", unimportance=2
            )
            refl.rebin(new_q=spacing(refl.q_vectors.min(), refl.q_vectors.max(),
                                      the_boss.data.q_step))
        the_boss.reduction.data_state.rebinned = the_boss.data.q_shape

    the_boss.data_source.experiment.measurement.q_range = [
        str(refl.q_vectors.min()), str(refl.q_vectors.max())]
    the_boss.data.n_qvectors = str(len(refl.reflectivity))



    # Prepare the data array.
    if the_boss.reduction.normalisation==True:
        data = np.array([refl.q_vectors, refl.reflectivity, refl.reflectivity_e]).T
    elif the_boss.reduction.normalisation==False:
        data = np.array([refl.q_vectors, refl.reflectivity_nonorm, refl.reflectivity_e_nonorm]).T
    debug.log("XRR reduction completed.", unimportance=2)

    # Work out where to save the file.
    datetime_str = datetime.now().strftime("%Y-%m-%d_%Hh%Mm%Ss")
    run_numbers.sort()
    dat_filename = 'XRR_{}_'.format(
        run_numbers[0]) + yaml_pipeline_name + datetime_str + ".dat"
    if filename is None:
        # Make sure that the processing directory exists.
        processing_path = path.join(the_boss.directory_path, 'processing')
        if not os.path.exists(processing_path):
            os.makedirs(processing_path)
        # Now prepare the full path to the file
        filename = (processing_path + dat_filename)
    if str(filename).endswith('processed'):
        if not os.path.exists(str(filename)):
            os.makedirs(str(filename)+'/')
    if os.path.isdir(str(filename)):
        # It's possible we were given a directory in which to save the created
        # file. In this case, use the filename variable as a directory and add
        # our auto generated filename to it.
        filename = os.path.join(str(filename), dat_filename)

    # Write the data.
    np.savetxt(
        filename, data, header=f"{dump(vars(the_boss))}\n Q(1/Å)\tR\tR_error"
    )

    debug.log("-" * 10)
    debug.log(f"Reduced data stored at {filename}", unimportance=0)
    debug.log("-" * 10)



"""
A profile is a measurement resulting from a scan, or a series of scans. Profiles
are the central objects in the islatu library, containing the total reflected
intensity as a function of scattering vector data.
"""

from typing import List

from islatu.scan import Scan
from islatu.stitching import concatenate, rebin
from islatu.data import Data


class Profile(Data):
    """
    The object that is used to store all information relating to a reflectivity
    profile.
    """

    def __init__(self, data: Data, scans: List[Scan]) -> None:
        super().__init__(data.intensity, data.intensity_e, data.energy,
                         data.theta)
        self.scans = scans

    @classmethod
    def fromfilenames(cls, filenames, parser,adjustments=None):
        """
        Instantiate a profile from a list of scan filenames.

        Args:
            filenames (:py:attr:`list`):
                List of files, one for each reflectometry scan. Can have length
                one.
            parser (:py:attr:`callable`):
                Parser function for the reflectometry scan files.
        """

        # Load the scans, specifying the scan axis name if necessary.
        if adjustments==None:
            scans = [parser(filename) for filename in filenames]
        else:
            scans = [parser(filename,adjustments=adjustments) for filename in filenames]

        # Now that the individual scans have been loaded, data needs to be
        # constructed. The simplest way to do this is by concatenating the
        # data from each of the constituent scans.
        q_vectors, intensity, intensity_e = concatenate(scans)

        # Note: we are making the implicit assumption that energy is independent
        # of scan number at this point.
        energy = scans[0].metadata.probe_energy

        data = Data(intensity, intensity_e, energy, q_vectors=q_vectors)

        return cls(data, scans)

    def crop(self, crop_function, **kwargs):
        """
        Calls the Class method for the :func:`~islatu.scan.Scan2D.crop`
        method for each :py:class:`~Scan2D` in :py:attr:`self.scans`.

        Args:
            crop_function (:py:attr:`callable`): Cropping function to be used.
            kwargs (:py:attr:`dict`, optional): Keyword arguments for the
                cropping function. Defaults to :py:attr:`None`.
        """
        for scan in self.scans:
            scan.crop(crop_function, **kwargs)
        self.concatenate()

    def bkg_sub(self, bkg_sub_function, **kwargs):
        """
        Class method for the :func:`~islatu.refl_data.Scan.bkg_sub` method for
        each :py:class:`~Scan` in the list.

        Args:
            bkg_sub_function (:py:attr:`callable`): Background subtraction
                function to be used.
            kwargs (:py:attr:`dict`, optional): Keyword arguments for
                the background subtraction function. Defaults to
                :py:attr:`None`.
        """
        # When a scan subtracts background from each of its images, its
        # background subtraction function may expose information relating to the
        # subtraction process. This information will be stored in bkg_sub_info.
        bkg_sub_info = []
        # Now just iterate over all of the scans in the profile and subtract the
        # background, storing the return values in bkg_sub_info.
        for scan in self.scans:
            bkg_sub_info.append(scan.bkg_sub(
                bkg_sub_function, **kwargs))

        self.concatenate()

        # Expose the optimized fit parameters for meta-analysis.
        return bkg_sub_info

    def subsample_q(self, scan_identifier, q_min=0, q_max=float('inf')):
        """
        For the scan with identifier scan_identifier, delete all data points for
        which q < q_min or q > q_max.

        Args:
            scan_identifier:
                The scan ID of the scan to be subsampled. This must be a unique
                substring of the filename from which the scan was taken. For
                example, if a scan's nexus filename is i07-413244.nxs, then
                a valid scan_ID would be "413244", as this string will uniquely
                identify the correct scan from within the profile.
            q_min:
                The smallest acceptable value of q. Defaults to 0 Å.
            q_max:
                The largest acceptable value of q. Defaults to inf Å.
        """
        for scan in self.scans:
            print(scan_identifier, scan.metadata.src_path)
            if scan_identifier in scan.metadata.src_path:
                scan.subsample_q(q_min, q_max)
        self.concatenate()

    def footprint_correction(self, beam_width, sample_size):
        """
        Class method for :func:`~islatu.refl_data.Scan.footprint_correction`
        for each :py:class:`~Scan` in the list.

        Args:
            beam_width (:py:attr:`float`): Width of incident beam, in metres.
            sample_size (:py:class:`uncertainties.core.Variable`): Width of
                sample in the dimension of the beam, in metres.
            theta (:py:attr:`float`): Incident angle, in degrees.
        """
        for scan in self.scans:
            scan.footprint_correction(beam_width, sample_size)
        self.concatenate()

    def transmission_normalisation(self, overwrite_transmissions=None):
        """
        Perform the transmission correction.
        """
        for i, scan in enumerate(self.scans):
            if overwrite_transmissions is not None:
                overwrite_transmission = overwrite_transmissions[i]
            else:
                overwrite_transmission = None
            scan.transmission_normalisation(overwrite_transmission)

        self.concatenate()

    def qdcd_normalisation(self, itp):
        """
        Class method for :func:`~islatu.refl_data.Scan.qdcd_normalisation` for
        each :py:class:`~Scan` in the list.

        Args:
            normalisation_file (:py:attr:`str`): The ``.dat`` file that
                contains the normalisation data.
        """
        for scan in self.scans:
            scan.qdcd_normalisation(itp)
        self.concatenate()

    def concatenate(self):
        """
        Class method for :func:`~islatu.stitching.concatenate`.
        """
        self.q_vectors, self.intensity, self.intensity_e = \
            concatenate(self.scans)

    def rebin(self, new_q=None, rebin_as="linear", number_of_q_vectors=5000):
        """
        Class method for :func:`islatu.stitching.rebin`.

        Args:
            new_q (:py:attr:`array_like`):
                Array of potential q-values. Defaults to :py:attr:`None`. If
                this argument is not specified, then the new q, R values are
                binned according to rebin_as and number_of_q_vectors.
            rebin_as (py:attr:`str`):
                String specifying how the data should be rebinned. Options are
                "linear" and "log". This is only used if the new_q are
                unspecified.
            number_of_q_vectors (:py:attr:`int`, optional):
                The max number of q-vectors to be using initially in the
                rebinning of the data. Defaults to :py:attr:`400`.
        """
        self.q_vectors, self.intensity, self.intensity_e = rebin(
            self.q_vectors, (self.intensity, self.intensity_e), new_q,
            rebin_as=rebin_as, number_of_q_vectors=number_of_q_vectors)



"""
module to load in process configurations and check against preset schemas
"""
import os
import ast
from schema import Or, And, Use,Schema, SchemaError,Optional

import yaml
import numpy as np
import datetime





def validate_instrument(inst):
    """
    check instrument is from allowed types
    """
    valid_instruments=['i07']
    if inst not in valid_instruments:
        raise ValueError(f"instrument {inst} not in valid instrument list {valid_instruments}")
    return True

def validate_crop_method(method):
    """
    check crop method is from allowed types
    """
    valid_methods=['crop']
    if method not in valid_methods:
        raise ValueError(f"method {method} not in valid crop method list {valid_methods}")
    return True


def validate_background_method(method):
    """
    check background method is from allowed types
    """
    valid_methods=['roi_subtraction','None','none']
    if method not in valid_methods:
        raise ValueError(f"method {method} not in valid background method list {valid_methods}")
    return True

def validate_new_axis(name):
    """
    check new axis name is from allowed types
    """
    valid_new_axes=['diff1chi','diff1delta','diff2alpha']
    if name not in valid_new_axes:
        raise ValueError(f"axis name {name} not in valid new axis list {valid_new_axes}")
    return True

def validate_new_type(type):
    """
    check new axis type is from allowed types
    """
    valid_types=['th','tth','q']
    if type not in valid_types:
        raise ValueError(f" type {type} not in valid types list {valid_types}")
    return True

normalisation_schema=Schema({
    "maxnorm": bool,

})
transmission_schema=Schema({
    "values": lambda t: isinstance(t, list) and len(t) == 2 and all(isinstance(x, float) for x in t),
})

setup_schema=Schema({
    "sample size":And( Use(lambda s: ast.literal_eval(s)),  # Convert string to tuple
    lambda t: isinstance(t, tuple) and len(t) == 2 and all(isinstance(x, float) for x in t)),
    "beam width": And( Use(float), lambda f: isinstance(f, float)),
    Optional("dcd normalisation"): str,

})
visit_schema=Schema({
   "local contact":str,
   "user": str,
   "user affiliation": str,
   "visit id": str,
   'date': And(lambda d: isinstance(d, datetime.date))

})

region_schema1=Schema({
   "x_start": int,
   "x_end": int,
   "y_start": int,
   "y_end": int,
})

region_schema2=Schema({
   "x": int,
   "width": int,
   "y": int,
   "height": int,
})

crop_schema=Schema({
  "method": And(str,validate_crop_method),
  Optional("kwargs"): And(dict,Or(region_schema1,region_schema2)),

})

background_schema=Schema({
  "method": And(str,validate_background_method),
  Optional("kwargs"): And(dict,Or(region_schema1,region_schema2)),

})



adjust_schema=Schema({
    Optional("new_axis_name"):And(str,validate_new_axis),
    Optional("new_axis_type"): And(str,validate_new_type),
    Optional("theta_offset"):float,
    Optional("q_offset"):float,

})

rebin_schema=Schema({
    "n qvectors": int,

})

config_schema = Schema({
    "instrument": And(str,validate_instrument),
    "setup": setup_schema,
    "visit": visit_schema,
    Optional("crop"): crop_schema,
    Optional("background"): background_schema,
    Optional("adjustments"): adjust_schema,
    Optional("rebin"): rebin_schema ,
    Optional("normalisation"): normalisation_schema,
    Optional("transmission"): transmission_schema,
    Optional("output_columns"): int,

})


def check_config_schema(input_config: dict):
    """
    check a process configuration against define scheme
    """
    try:
        config_schema.validate(input_config)
        print("config values loaded")
        return True
    except SchemaError as se:
        raise se





"""
Background substraction is a necessary component of reflectometry reduction,
where the background scattering is removed from the reflected intensity.

Herein are some functions to enable that for a two-dimensional detector image,
as well as simple dataclasses in which we can store some information relating to
the background subtraction, and any fitting that we might have carried out.
"""

from dataclasses import dataclass
from typing import Callable, List

import numpy as np
from scipy.stats import norm
from scipy.optimize import curve_fit

from islatu.region import Region
from islatu.image import Image


@dataclass
class FitInfo:
    """
    A simple dataclass in which we can store data relating to the quality of a
    fit.
    """
    popt: np.ndarray
    pcov: np.ndarray
    fit_function: Callable


@dataclass
class BkgSubInfo:
    """
    A simple data class in which we can store information relating to a
    background subtraction.
    """
    bkg: float
    bkg_e: float
    bkg_sub_function: Callable
    fit_info: FitInfo = None


def roi_subtraction(image, list_of_regions: List[Region]):
    """
    Carry out background subtraction by taking a series of rectangular regions
    of interested (ROIs) as being fair Poissonian measurements of the
    background.

    Args:
        image:
            The islatu.image.Image object from which we should subtract
            background from.
        list_of_regions:
            A list of instances of the Regions class corresponding to background
            regions.
    """
    # We're going to need to count all intensity in all the background, as well
    # as the number of pixels used in our measurement of the background.
    sum_of_bkg_areas = 0
    total_num_pixels = 0

    # Make sure we've been given multiple regions. If not, np: make a list.
    if isinstance(list_of_regions, Region):
        list_of_regions = [list_of_regions]

    # Add up all the intensity in all the pixels.
    for region in list_of_regions:
        # Now add the total intensity in this particular background region to
        # the intensity measured in all the background regions so far.
        sum_of_bkg_areas += np.sum(
            image.array_original[
                int(region.y_start):int(region.y_end),
                int(region.x_start):int(region.x_end)
            ]
        )
        # Add the number of pixels in this background ROI to the total number of
        # pixels used to compute the background measurement overall.
        total_num_pixels += region.num_pixels

    # Now Poisson stats can be abused to only calculate a single sqrt.
    err_of_bkg_areas = np.sqrt(sum_of_bkg_areas)
    if err_of_bkg_areas == 0:
        err_of_bkg_areas = 1

    # Get the per pixel background mean and stddev.
    bkg_per_pixel = sum_of_bkg_areas / total_num_pixels
    bkg_error_per_pixel = err_of_bkg_areas / total_num_pixels

    # Expose the calculated background and background_error per pixel.
    return BkgSubInfo(bkg_per_pixel, bkg_error_per_pixel, roi_subtraction)


def univariate_normal(data, mean, sigma, offset, factor):
    """
    Produce a univariate normal distribution.

    Args:
        data (:py:attr:`array_like`): Abscissa data.
        mean (:py:attr:`float`): Mean (horizontal).
        sigma (:py:attr:`float`): Variance (horizontal).
        offset (:py:attr:`float`): Offset from the 0 for the ordinate, this is\
            the background level.
        factor (:py:attr:`float`): Multiplicative factor for area of normal\
            distribution.

    Returns:
        :py:attr:`array_like`: Ordinate data for univariate normal distribution.
    """
    # Creation of the bivariate normal distribution
    normal = norm(loc=mean, scale=sigma)
    return offset + normal.pdf(data).flatten() * factor


def fit_gaussian_1d(image: Image, params_0=None, bounds=None, axis=0):
    """
    Fit a one-dimensional Gaussian function with some ordinate offset to an
    image with uncertainty. This is achieved by averaging in a given ``axis``
    before performing the fit. Return the results, and index of the offset.

    Args:
        image:
            The islatu image object to fit.
        params_0 (:py:attr:`list`, optional):
            An initial guess at the parameters. Defaults to values based on the
            image.
        bounds (:py:attr:`list` of :py:attr:`tuple`, optional):
            Bounds for the fitting. Defaults to values based on the image.
        axis (:py:attr:`int`):
            The dimension along which the averaging will be performed.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`array_like`: The results (with uncertainties) for each
                of the 6 parameters fit.
            - :py:attr:`int`: The index of the offset.
            - :py:attr:`None`: As it is not possible to describe the reflected
                peak width.
    """
    arr, arr_e = image.array, image.array_e
    ordinate = arr.mean(axis=axis)

    # Now we can generate an array of errors.
    ordinate_e = np.sqrt(np.mean(arr_e**2, axis=axis))

    # Setting default values.
    if params_0 is None:
        # Now we generate the initial values for our Gaussian fit.
        # These values are crucial – as this is a high dimensional fitting
        # problem, it is likely that we'll get stuck in a local minimum if these
        # aren't good.
        # Guess that the Gaussian mean is at the most intense mean pixel value.
        mean0 = np.argmax(ordinate)
        # Guess that the standard deviation is a single pixel.
        sdev0 = 1
        # Guess that the background (offset) is the median pixel value.
        offset0 = np.median(ordinate)
        # Guess that the scale is equal to the largest recorded value.
        scale0 = arr.max()
        params_0 = [mean0, sdev0, offset0, scale0]
    if bounds is None:
        bounds = ([0, 0, 0, 0],
                  [ordinate.shape[0], ordinate.shape[0], scale0, scale0 * 10])

    # Perform the fitting.
    fit_popt_pcov = curve_fit(
        univariate_normal,
        np.arange(0, ordinate.shape[0], 1), ordinate, bounds=bounds,
        sigma=ordinate_e, p0=params_0, maxfev=2000 * (len(params_0) + 1))

    fit_info = FitInfo(fit_popt_pcov[0], fit_popt_pcov[1], univariate_normal)

    # Determine uncertainty from covarience matrix.
    # Note: the stddev of the fit Gaussian can be accessed via popt[1].
    p_sigma = np.sqrt(np.diag(fit_info.pcov))

    return BkgSubInfo(fit_info.popt[2], p_sigma[2], fit_gaussian_1d, fit_info)



"""
Reflectometry data must be corrected as a part of reduction.
These functions facilitate this, including the footprint and
DCD q-variance corrections.
"""


import numpy as np
from scipy.stats import norm
from scipy.interpolate import splrep


def footprint_correction(beam_width, sample_size, theta):
    """
    The factor by which the intensity should be multiplied to account for the scattering geometry, where the beam is Gaussian in shape.

    Args:
        beam_width (:py:attr:`float`):
            Width of incident beam, in metres.
        sample_size (:py:attr:`float`):
            Width of sample in the dimension of the beam, in metres.
        theta (:py:attr:`float`):
            Incident angle, in degrees.

    Returns:
        Array of correction factors.
    """
    # Deal with the [trivial] theta=0 case.
    theta = np.array([10**(-3) if t == 0 else t for t in theta])

    beam_sd = beam_width / 2 / np.sqrt(2 * np.log(2))
    projected_beam_sd = beam_sd / np.sin(np.radians(theta))
    frac_of_beam_sampled = (
        norm.cdf(sample_size/2, 0, projected_beam_sd) -
        norm.cdf(-sample_size/2, 0, projected_beam_sd)
    )
    return frac_of_beam_sampled


def get_interpolator(
        file_path, parser, q_axis_name="qdcd_", intensity_axis_name="adc2"):
    """
    Get an interpolator object from scipy, this is useful for the DCD q-normalisation step.

    Args:
        file_path (:py:attr:`str`): File path to the normalisation file.
        parser (:py:attr:`callable`): Parser function for the normalisation file.
        q_axis_name (:py:attr:`str`, optional): Label for the q-value in the normalisation file. Defaults to ``'qdcd_'``.
        intensity_axis_name (:py:attr:`str`, optional): Label for the intensity in the normalisation file. Defaults to ``'adc2'``.

    Returns:
        :py:attr:`tuple`: Containing:
            - :py:attr:`array_like`: Interpolation knots.
            - :py:attr:`array_like`: B-spline coefficients.
            - :py:attr:`int`: Degree of spline.
    """
    normalisation_data = parser(file_path)[1].sort_values(by='qdcd_')
    return splrep(
        normalisation_data[q_axis_name],
        normalisation_data[intensity_axis_name])



"""
Often the detector is a lot larger than the reflected intensity peak, so it
makes sense to crop the image to the peak.
"""


import numpy as np

from islatu.region import Region


def crop_to_region(array: np.ndarray, region: Region):
    """
    Crops the input array to the input region.

    Args:
        array:
            The array to crop.
        region:
            The instance of Region to crop to.
    """
    return array[region.y_start:region.y_end,region.x_start:region.x_end]



import io
from glob import glob
from os.path import basename, dirname, join, splitext, abspath

from setuptools import find_packages
from setuptools import setup


THIS_DIRECTORY = abspath(dirname(__file__))
with io.open(join(THIS_DIRECTORY, 'README.md')) as f:
    LONG_DESCRIPTION = f.read()

REQUIREMENTS = [
    "wheel",
    "numpy",
    "scipy",
    "coverage",
    "pandas",
    "pyyaml",
    "nexusformat",
    "pytest",
    "pytest-lazy-fixture",
    "nbsphinx",
    "jupyter-sphinx",
    "jupyterlab",
    "ipywidgets",
    "pytest-cov",
]

setup(
    name='islatu',
    version='1.3.1',
    license='MIT',
    description='A package for the reduction of reflectometry data.',
    author='Philip Mousley',
    author_email='philip.mousley@diamond.ac.uk',
    long_description=LONG_DESCRIPTION,
    long_decription_content_type='text/markdown',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    setup_requires=REQUIREMENTS,
    install_requires=REQUIREMENTS
)



"""
This module contains the Scan and Scan2D classes. A Scan is a measurement and so
inherits from MeasurementBase. An instance of Scan contains scan metadata, as
well as a suite of methods useful for data correction, uncertainty calculations
and the like.

A Scan2D is a Scan whose Data object's intensity values are computed from an
image captured by an area detector. Many of Scan's methods are overloaded to
make use of the additional information provided by the area detector, and extra
image manipulation methods are included in Scan2D.
"""

from typing import List

import numpy as np
from scipy.interpolate import splev

import islatu.corrections as corrections
from islatu.metadata import Metadata
from islatu.data import Data, MeasurementBase
from islatu.image import Image


class Scan(MeasurementBase):
    """
    A class used to store reflectometry scans taken with a point detector.
    """

    def __init__(self, data: Data, metadata: Metadata) -> None:
        # Initialize the MeasurementBase from Data. This is much simpler than
        # passing a million arguments directly to the scan.
        super().__init__(data.intensity, data.intensity_e, data.energy,
                         metadata, data.theta)

    def subsample_q(self, q_min=0, q_max=float('inf')):
        """
        Delete data points less than q_min and more than q_max.

        Args:
            q_min:
                The minimum q to be included in this scan. Defaults to 0 Å.
            q_max:
                The maximum q to be included in this scan. Defaults to inf Å.
        """
        # A place to store all the indices violating our condition on q.
        illegal_q_indices = np.where(
            (self.q_vectors <= q_min) | (self.q_vectors >= q_max)
        )[0]
        # [0] necessary because np.where returns a tuple of arrays of length 1.
        # This is a quirk of np.where – I don't think it's actually designed to
        # be used like this, and they encourage np.asarray(condition).nonzero()

        # Now remove all data points at these qs.
        self.remove_data_points(illegal_q_indices)
    
    def transmission_normalisation(self,overwrite_transmission=None):
        """
        Perform the transmission correction.
        """
        
        if overwrite_transmission is not None:
            transmissionvalue=overwrite_transmission
        elif 'transmissions' in dir(self.metadata):
            transmissionvalue=self.metadata.transmissions
        else:
            transmissionvalue=self.metadata.transmission
            
        if np.size(np.array(transmissionvalue))==1:
            self.intensity /= float(transmissionvalue)
            self.intensity_e /= float(transmissionvalue)
        else:
            self.intensity /= transmissionvalue
            self.intensity_e /= transmissionvalue

    def qdcd_normalisation(self, itp):
        """
        Perform normalisation by DCD variance.

        Args:
            itp (:py:attr:`tuple`): Containing interpolation knots
                (:py:attr:`array_like`), B-spline coefficients
                (:py:attr:`array_like`), and degree of spline (:py:attr:`int`).
        """
        self.intensity /= splev(self.q_vectors, itp)
        self.intensity_e /= splev(self.q_vectors, itp)

    def footprint_correction(self, beam_width, sample_size):
        """
        Class method for :func:`islatu.corrections.footprint_correction`.

        Args:
            beam_width (:py:attr:`float`): Width of incident beam, in metres.
            sample_size (:py:class:`uncertainties.core.Variable`): Width of
                sample in the dimension of the beam, in metres.
            theta (:py:attr:`float`): Incident angle, in degrees.
        """
        frac_of_beam_sampled = corrections.footprint_correction(
            beam_width, sample_size, self.theta)
        self.intensity /= frac_of_beam_sampled
        self.intensity_e /= frac_of_beam_sampled


class Scan2D(Scan):
    """
    Attributes:
        data (:py:attr:`islatu.data.Data`):
            The intensity as a function of Q data for this scan.
        metadata (:py:attr:`islatu.metadata.Metadata`):
            This scan's metadata.
        images (:py:attr:`list` of :py:class:`islatu.image.Image`):
            The detector images in the given scan.
    """

    def __init__(self, data: Data, metadata: Metadata, images: List[Image],remove_indices=None) \
            -> None:
        super().__init__(data, metadata)
        self.images = images
        self.detname=self.metadata.detector_name
        if 'attenuation_filters_moving' in self.metadata.entry[f'{self.detname}'].keys():
            try:
                filterslist=self.metadata.entry[f'{self.detname}/attenuation_filters_moving'].nxdata
            except (AttributeError,TypeError):
                filterslist=[]
            if len(filterslist) >1:
                self.metadata.transmissionsraw=self.metadata.entry[f'{self.detname}_transmission/transmission'].nxdata
                self.metadata.transmissions=np.delete(np.insert(self.metadata.transmissionsraw,0,1e-9),-1)
        if remove_indices is not None:
            self.remove_data_points(remove_indices)
            
    def crop(self, crop_function, **kwargs):
        """
        Crop every image in images according to crop_function.

        args:
            crop_function (:py:attr:`callable`):
                Cropping function to be used.
            kwargs (:py:attr:`dict`, optional):
                Keyword arguments for the cropping function. Defaults to
                :py:attr:`None`.
            progress (:py:attr:`bool`, optional):
                Show a progress bar. Requires the :py:mod:`tqdm` package.
                Defaults to :py:attr:`True`.
        """

        (vals, stdevs) = (np.zeros(len(self.intensity)),
                          np.zeros(len(self.intensity)))
        for i, image in enumerate(self.images):
            image.crop(crop_function, **kwargs)
            vals[i], stdevs[i] = self.images[i].sum()

        self.intensity = np.array(vals)
        self.intensity_e = np.array(stdevs)

    def bkg_sub(self, bkg_sub_function, **kwargs):
        """
        Perform background substraction for each image in a Scan.

        Args:
            bkg_sub_function (:py:attr:`callable`): Background subtraction
                function to be used.
            kwargs (:py:attr:`dict`, optional): Keyword arguments for
                the background subtraction function. Defaults
                to :py:attr:`None`.
            progress (:py:attr:`bool`, optional): Show a progress bar.
                Requires the :py:mod:`tqdm` package. Defaults
                to :py:attr:`True`.
        """
        vals, stdevs = np.zeros(
            len(self.intensity)), np.zeros(len(self.intensity))

        # We keep track of the bkg_sub_infos for meta-analyses.
        bkg_sub_info = [
            image.background_subtraction(bkg_sub_function, **kwargs)
            for image in self.images
        ]

        # Also update the image intensities & errors.
        for i, image in enumerate(self.images):
            vals[i], stdevs[i] = image.sum()

        # Store the intensity(Q) to the new value.
        self.intensity = np.array(vals)
        self.intensity_e = np.array(stdevs)

        # Expose the information relating to the background subtraction.
        return bkg_sub_info

    def remove_data_points(self, indices):
        """
        Convenience method for the removal of specific data points by their
        indices.

        Args:
            indices:
                The indices to be removed.
        """
        super().remove_data_points(indices)
        if 'transmissions' in dir(self.metadata):
            self.metadata.transmissions =np.delete(self.metadata.transmissions,indices)

        # Delete images in reverse order if you don't like errors.
        for idx in sorted(indices, reverse=True):
            del self.images[idx]



"""
This module contains the Metadata class, returned by parser methods in the
islatu.io module. This class provides a consistent way to refer to metadata
returned by different detectors/instruments, and also contains a dictionary
of all of the metadata as scraped from the parsed file.
"""

from abc import abstractmethod

import numpy as np


class Metadata:
    """
    An ABC for classes that store metadata parsed from data files. This defines
    the properties that must be implemented by parsing classes.
    """

    def __init__(self, local_path):
        self.local_path = local_path

    @property
    @abstractmethod
    def probe_energy(self):
        """
        This must be overridden.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def default_axis(self) -> np.ndarray:
        """
        Returns a numpy array of data associated with the default axis, where
        "default axis" should be understood in the NeXus sense to mean the
        experiment's dependent variable.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def default_axis_name(self) -> str:
        """
        Returns the name of the default axis, as it was recorded in the data
        file stored at local_path.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def default_axis_type(self) -> str:
        """
        Returns what type of default axis we have. Options are 'q', 'th' or
        'tth'.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def transmission(self):
        """
        Proportional to the fraction of probe particles allowed by an attenuator
        to strike the sample.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def detector_distance(self):
        """
        Returns the distance between sample and detector.
        """
        raise NotImplementedError()
